import logging
from math import sqrt
import os
from hepdata_converter.common import Option
from hepdata_converter.writers import Writer
import abc

from hepdata_converter.writers.utils import error_value_processor

logging.basicConfig()
log = logging.getLogger(__name__)

class ObjectWrapper(object):
    __metaclass__ = abc.ABCMeta
    accept_alphanumeric = True
    # :core_object: bool if set to false object of this class will be created as an *extra* additionally to the first
    # object of class with ```core_object``` set to True
    core_object = True

    @classmethod
    def is_number_var(cls, *variables):
        is_number_list = []
        for variable in variables:
            for element in variable['values']:
                if 'value' in element and isinstance(element['value'], (str, unicode)):
                    is_number_list.append(False)
                else:
                    is_number_list.append(True)
        return is_number_list

    @classmethod
    def sanitize_name(cls, name):
        return name.replace(' ', '_').replace('/', '_')

    @classmethod
    def is_value_var(cls, variable):
        return cls._is_attr_variable('value', variable)

    @classmethod
    def is_range_var(cls, variable):
        return cls._is_attr_variable('high', variable) and cls._is_attr_variable('low', variable)

    @classmethod
    def has_errors(cls, variable):
        return cls._is_attr_variable('errors', variable)

    @classmethod
    def _is_attr_variable(cls, attr_name, variable):
        for element in variable['values']:
            if attr_name not in element:
                return False
        return True

    def __init__(self, independent_variable_map, dependent_variable, dependent_variable_index):
        self.xval = []
        self.yval = []
        self.xerr_minus = []
        self.xerr_plus = []
        self.yerr_minus = []
        self.yerr_plus = []
        self.independent_variables = list(independent_variable_map)
        self.independent_variable_map = independent_variable_map
        self.dependent_variable = dependent_variable
        self.dependent_variable_index = dependent_variable_index

    @classmethod
    def match(cls, independent_variables_map, dependent_variable):
        if not cls.accept_alphanumeric:
            # don't match if any independent or dependent variable values are non-numeric
            for var in independent_variables_map:
                if False in cls.is_number_var(dependent_variable):
                    return False
            if False in cls.is_number_var(dependent_variable):
                return False
        return True

    @classmethod
    def match_and_create(cls, independent_variables_map, dependent_variable, dependent_variable_index):
        if cls.match(independent_variables_map, dependent_variable):
            return cls(independent_variables_map, dependent_variable, dependent_variable_index).create_objects()
        return []

    def calculate_total_errors(self):
        is_number_list = self.is_number_var(self.dependent_variable)
        for independent_variable in self.independent_variable_map:
            xerr_minus = []
            self.xerr_minus.append(xerr_minus)
            xerr_plus = []
            self.xerr_plus.append(xerr_plus)
            xval = []
            self.xval.append(xval)
            ArrayWriter.calculate_total_errors(independent_variable, is_number_list,
                                               xerr_minus, xerr_plus, xval)
        ArrayWriter.calculate_total_errors(self.dependent_variable, is_number_list,
                                           self.yerr_minus, self.yerr_plus, self.yval)

    @abc.abstractmethod
    def create_objects(self):
        pass


class ObjectFactory(object):
    def __init__(self, class_list, independent_variables, dependent_variables):
        self.class_list = class_list
        self.map = {}
        self.independent_variables = independent_variables
        self.dependent_variables = dependent_variables
        for variable_index in xrange(len(dependent_variables)):
            self.map[variable_index] = list(independent_variables)

    def get_next_object(self):
        for dependent_variable_index in xrange(len(self.dependent_variables)):
            auxiliary_object_created = False
            for class_wrapper in self.class_list:
                # if auxiliary object was already created for this variable, don't create more
                # (eg. after creating TH3 (which does not consume variables) don't create TH2 & TH1,
                # but go for best fit of "core" objects eg. TGraph

                if not class_wrapper.core_object and auxiliary_object_created:
                    continue

                objects = class_wrapper.match_and_create(self.map[dependent_variable_index],
                                                         self.dependent_variables[dependent_variable_index], dependent_variable_index)
                if objects and not class_wrapper.core_object:
                    auxiliary_object_created = True

                for obj in objects:
                    yield obj


class ArrayWriter(Writer):
    __metaclass__ = abc.ABCMeta

    @staticmethod
    def calculate_total_errors(variable, is_number_list, min_errs, max_errs, values):
        for i, entry in enumerate(variable['values']):
            if not is_number_list[i]: continue # skip non-numeric y values
            if 'value' in entry:
                values.append(entry['value'])
                if 'errors' in entry:
                    errors_min = 0.0
                    errors_max = 0.0
                    for error in entry['errors']:
                        if 'asymerror' in error:
                            err_minus = error_value_processor(entry['value'], error['asymerror']['minus'])
                            err_plus = error_value_processor(entry['value'], error['asymerror']['plus'])
                            errors_min += pow(min(err_plus, err_minus, 0.0), 2)
                            errors_max += pow(max(err_plus, err_minus, 0.0), 2)
                        elif 'symerror' in error:
                            try:
                                err = error_value_processor(entry['value'], error['symerror'])
                                errors_min += pow(err, 2)
                                errors_max += pow(err, 2)
                            except TypeError:
                                print log.error('TypeError encountered when parsing {0}'.format(error['symerror']))

                    min_errs.append(sqrt(errors_min))
                    max_errs.append(sqrt(errors_max))
                elif 'low' in entry and 'high' in entry:
                    min_errs.append(entry['value'] - entry['low'])
                    max_errs.append(entry['high'] - entry['value'])
                else:
                    min_errs.append(0.0)
                    max_errs.append(0.0)
            else:
                middle_val = (entry['high'] - entry['low']) * 0.5 + entry['low']
                values.append(middle_val)
                min_errs.append(middle_val - entry['low'])
                max_errs.append(entry['high'] - middle_val)

    @classmethod
    def options(cls):
        options = super(ArrayWriter, cls).options()
        options['table'] = Option('table', 't', required=False, variable_mapping='table_id', default=None,
                                  help=('Specifies which table should be exported, if not specified all tables will be exported '
                                        '(in this case output must be a directory, not a file)'))

        return options

    def __init__(self, *args, **kwargs):
        kwargs['single_file_output'] = True
        super(ArrayWriter, self).__init__(*args, **kwargs)
        self.tables = []
        self.extension = None

    @abc.abstractmethod
    def _write_table(self, data_out, table):
        pass

    def _get_tables(self, data_in):
        # get table to work on
        if self.table_id is not None:
            if isinstance(self.table_id, int):
                self.tables.append(data_in.get_table(id=self.table_id))
            else:
                try:
                    tab = data_in.get_table(file=self.table_id)
                except IndexError:
                    tab = data_in.get_table(name=self.table_id)

                self.tables.append(tab)
        else:
            self.tables = data_in.tables

    def _prepare_outputs(self, data_out, outputs):
        if isinstance(data_out, (str, unicode)):
            self.file_emulation = True
            if self.table_id is not None:
                f = open(data_out, 'w')
                outputs.append(f)
            # data_out is a directory
            else:
                # create output dir if it doesn't exist
                self.create_dir(data_out)
                for table in self.tables:
                    outputs.append(open(os.path.join(data_out, table.name.replace(' ','').replace('/','-') + '.' + self.extension), 'w'))
        # multiple tables - require directory
        elif len(self.tables) > 1 and not isinstance(data_out, (str, unicode)):
            raise ValueError("Multiple tables, output must be a directory")
        else:
            outputs.append(data_out)

    def write(self, data_in, data_out, *args, **kwargs):
        """

        :param data_in:
        :type data_in: hepconverter.parsers.ParsedData
        :param data_out: filelike object
        :type data_out: file
        :param args:
        :param kwargs:
        """
        self._get_tables(data_in)

        self.file_emulation = False
        outputs = []

        self._prepare_outputs(data_out, outputs)

        for i in xrange(len(self.tables)):
            data_out = outputs[i]
            table = self.tables[i]

            self._write_table(data_out, table)

            if self.file_emulation:
                data_out.close()

    @classmethod
    def _extract_independent_variables(cls, table, headers, data, qualifiers_marks):
        for independent_variable in table.independent_variables:
            name = independent_variable['header']['name']
            if 'units' in independent_variable['header']:
                name += ' [%s]' % independent_variable['header']['units']
            headers.append(unicode(name).encode('utf8', 'replace'))
            x_data = []
            x_data_low = []
            x_data_high = []
            for value in independent_variable['values']:

                if 'high' in value and 'low' in value:
                    x_data_low.append(value['low'])
                    x_data_high.append(value['high'])
                    if 'value' in value:
                        x_data.append(value['value'])
                    else:
                        x_data.append(0.5*(value['low'] + value['high']))
                else:
                    x_data_low.append(value['value'])
                    x_data_high.append(value['value'])
                    x_data.append(value['value'])

            data.append(x_data)
            if x_data_high != x_data_low:
                data.append(x_data_low)
                data.append(x_data_high)
                header = headers[-1]
                headers.append(header + ' LOW')
                qualifiers_marks.append(False)
                headers.append(header + ' HIGH')
                qualifiers_marks.append(False)

    @classmethod
    def _parse_dependent_variable(cls, dependent_variable, headers, qualifiers, qualifiers_marks, data):
        units = ''
        if 'units' in dependent_variable['header']:
            units = ' [%s]' % dependent_variable['header']['units']
        headers.append(unicode(dependent_variable['header']['name'] + units).encode('utf8', 'replace'))

        qualifiers_marks.append(True)

        # peek at first value and create empty lists
        y_order = []
        y_data = {'values': []}
        y_order.append(y_data['values'])

        for value in dependent_variable['values']:

            # process the labels to ensure uniqueness
            observed_error_labels = {}
            for error in value.get('errors', []):
                label = error.get('label', 'error')

                if label not in observed_error_labels:
                    observed_error_labels[label] = 0
                observed_error_labels[label] += 1

                if observed_error_labels[label] > 1:
                    error['label'] = label + '_' + str(observed_error_labels[label])

                # append "_1" to first error label that has a duplicate
                if observed_error_labels[label] == 2:
                    for error1 in value.get('errors', []):
                        error1_label = error1.get('label', 'error')
                        if error1_label == label:
                            error1['label'] = label + "_1"
                            break

            # fill error template for all values
            for error in value.get('errors', []):
                label = error.get('label', 'error')
                if label + '_plus' not in y_data:
                    headers.append(label + ' +')
                    qualifiers_marks.append(False)
                    headers.append(label + ' -')
                    qualifiers_marks.append(False)
                    plus = []
                    y_data[label + '_plus'] = plus
                    y_order.append(plus)
                    minus = []
                    y_data[label + '_minus'] = minus
                    y_order.append(minus)

        for value in dependent_variable['values']:
            y_data['values'].append(value['value'])
            if 'errors' not in value:
                for key, val in y_data.items():
                    if key != 'values':
                        val.append(0)
            else:
                for key, val in y_data.items():
                    has_error = False
                    for i in xrange(len(value.get('errors', []))):
                        error = value['errors'][i]
                        label = error.get('label', 'error')

                        if 'symerror' in error:
                            error_plus = error['symerror']
                            error_minus = '-' + error_plus if isinstance(error_plus, (str, unicode)) else -error_plus
                        else:
                            error_plus = error['asymerror']['plus']
                            error_minus = error['asymerror']['minus']

                        if key == label + '_plus':
                            val.append(error_plus)
                            has_error = True
                        elif key == label + '_minus':
                            val.append(error_minus)
                            has_error = True

                    if key != 'values' and not has_error:
                        val.append(0)

        for entry in y_order:
            data.append(entry)

        for qualifier in dependent_variable.get('qualifiers', []):
            units = ''
            if 'units' in qualifier:
                units = ' [%s]' % qualifier['units']
            name = qualifier['name'] + units
            if name not in qualifiers:
                qualifiers[name] = []
            qualifiers[name].append(qualifier['value'])




def error_value_processor(value, error):
    """
    If an error is a percentage, we convert to a float, then
    calculate the percentage of the supplied value.

    :param value: base value, e.g. 10
    :param error: e.g. 20.0%
    :return: the absolute error, e.g. 12 for the above case.
    """
    if type(error) is str:
        if "%" in error:
            error = error.replace("%", "")
            error = float(error)

            error_abs = (value/100) * error
            return error_abs

    return error


# -*- encoding: utf-8 -*-
import hepdata_converter
from hepdata_converter.testsuite import insert_path, insert_data_as_str
from hepdata_converter.testsuite.test_writer import WriterTestSuite

__author__ = 'Michał Szostak'


class YAMLWriterTestSuite(WriterTestSuite):
    @insert_path('yaml_qual')
    @insert_data_as_str('csv/table_5_noqual.csv')
    def test_no_qal_parse(self, yaml_path, table_5_noqual):
        data = hepdata_converter.convert(yaml_path, options={'input_format': 'yaml',
                                                             'output_format': 'csv',
                                                             'single_file': True})

        self.assertEqual(data, table_5_noqual)



import inspect
import os
import pkgutil
from hepdata_converter.common import GetConcreteSubclassMixin, OptionInitMixin

__all__ = []

import abc


class Writer(GetConcreteSubclassMixin, OptionInitMixin):
    __metaclass__  = abc.ABCMeta

    def __init__(self, single_file_output, *args, **kwargs):
        OptionInitMixin.__init__(self, options=kwargs)
        self.single_file_output = single_file_output
        self.hepdata_doi = kwargs['hepdata_doi'].encode('utf8', 'replace') if 'hepdata_doi' in kwargs else ''

    @classmethod
    def create_dir(cls, path):
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != 17:
                raise

    @abc.abstractmethod
    def write(self, data_in, data_out, *args, **kwargs):
        """Writes data provided by data_in argument to data_out, data_out is implementation dependent, and can be
        directory, filelike object, etc

        :param data_in: input data (type is implementation specified)
        :param data_out: output data (type is implementation specified)
        :param args: additional arguments passed to concrete class
        :param kwargs: additional arguments passed to concrete class
        """


# import all packages in the parsers package, so that Writer.get_specific_writer will recognise them
for loader, name, is_pkg in pkgutil.walk_packages(__path__):
    module = loader.find_module(name).load_module(name)

    for name, value in inspect.getmembers(module):
        if name.startswith('__'):
            continue

        globals()[name] = value
        __all__.append(name)



# -*- coding: utf-8 -*-
import abc
from hepdata_converter.writers.array_writer import ArrayWriter, ObjectWrapper, ObjectFactory
import ROOT as ROOTModule
import array
import tempfile
import os
from hepdata_converter.writers.utils import error_value_processor

__author__ = 'Michał Szostak'

import logging
logging.basicConfig()
log = logging.getLogger(__name__)

class THFRootClass(ObjectWrapper):
    __metaclass__ = abc.ABCMeta

    _hist_axes_names = ['x', 'y', 'z']
    _hist_axes_getters = ['GetXaxis', 'GetYaxis', 'GetZaxis']
    dim = 0
    core_object = False

    def get_hist_classes(self):
        return [ROOTModule.TH1F, ROOTModule.TH2F, ROOTModule.TH3F]

    @classmethod
    def match(cls, independent_variables_map, dependent_variable):
        if not ObjectWrapper.match(independent_variables_map, dependent_variable):
            return False
        elif len(independent_variables_map) == cls.dim:
            for independent_variable in independent_variables_map:
                if not independent_variable['values']:
                    return False
                if 'low' not in independent_variable['values'][0] or 'high' not in independent_variable['values'][0]:
                    return False
            return True
        return False

    def _create_empty_hist(self, dependent_var_title, index, yval):

        xval = []
        for i in xrange(self.dim):
            xval.append([])
            i_var = self.independent_variables[i]['values']
            for x in i_var:
                xval[i].append(x['low'])
            xval[i].append(i_var[-1]['high'])

        name = "Hist%sD_y%s_e%s" % (self.dim, self.dependent_variable_index + 1, index)

        # order bin values of independent variables
        xval_ordered = []
        for i in xrange(self.dim):
            xval_ordered.append([])
            for j, x in enumerate(xval[i]):
                if j == 0:
                    x_highest = x
                    xval_ordered[i].append(x)
                else:
                    if x > x_highest:
                        x_highest = x
                        xval_ordered[i].append(x)

        if 1 == self.dim:
            nbinsx = len(xval_ordered[0]) - 1
            binsx = array.array('d', xval_ordered[0])
            hist = self.get_hist_classes()[self.dim - 1](self.sanitize_name(name), '', nbinsx, binsx)

        if 2 == self.dim:
            nbinsx = len(xval_ordered[0]) - 1
            binsx = array.array('d', xval_ordered[0])
            nbinsy = len(xval_ordered[1]) - 1
            binsy = array.array('d', xval_ordered[1])
            hist = self.get_hist_classes()[self.dim - 1](self.sanitize_name(name), '', nbinsx, binsx, nbinsy, binsy)

        if 3 == self.dim:
            nbinsx = len(xval_ordered[0]) - 1
            binsx = array.array('d', xval_ordered[0])
            nbinsy = len(xval_ordered[1]) - 1
            binsy = array.array('d', xval_ordered[1])
            nbinsz = len(xval_ordered[2]) - 1
            binsz = array.array('d', xval_ordered[2])
            hist = self.get_hist_classes()[self.dim - 1](self.sanitize_name(name), '', nbinsx, binsx, nbinsy, binsy, nbinsz, binsz)

        for i in xrange(self.dim):
            name = self.independent_variables[i]['header']['name']
            if 'units' in self.independent_variables[i]['header']:
                name += ' [%s]' % self.independent_variables[i]['header']['units']
            name = unicode(name).encode('ascii', 'replace')
            getattr(hist, self._hist_axes_getters[i])().SetTitle(name)
            if 'labels' in self.independent_variables[i]:
                for ibin, label in enumerate(self.independent_variables[i]['labels']):
                    getattr(hist, self._hist_axes_getters[i])().SetBinLabel(ibin + 1, label)

        if self.dim < len(self.get_hist_classes()):
            getattr(hist, self._hist_axes_getters[self.dim])().SetTitle(self.sanitize_name(dependent_var_title))

        for i in xrange(len(yval)):
            hist.Fill(*([self.xval[dim_i][i] for dim_i in xrange(self.dim)] + [yval[i]]))

        return hist

    def _create_hist(self, xval):

        name = "Hist%sD_y%s" % (self.dim, self.dependent_variable_index + 1)
        args = []

        # order bin values of independent variables
        xval_ordered = []
        for i in xrange(self.dim):
            xval_ordered.append([])
            for j, x in enumerate(xval[i]):
                if j == 0:
                    x_highest = x
                    xval_ordered[i].append(x)
                else:
                    if x > x_highest:
                        x_highest = x
                        xval_ordered[i].append(x)

        if 1 == self.dim:
            nbinsx = len(xval_ordered[0]) - 1
            binsx = array.array('d', xval_ordered[0])
            hist = self.get_hist_classes()[self.dim - 1](self.sanitize_name(name), '', nbinsx, binsx)

        if 2 == self.dim:
            nbinsx = len(xval_ordered[0]) - 1
            binsx = array.array('d', xval_ordered[0])
            nbinsy = len(xval_ordered[1]) - 1
            binsy = array.array('d', xval_ordered[1])
            hist = self.get_hist_classes()[self.dim - 1](self.sanitize_name(name), '', nbinsx, binsx, nbinsy, binsy)

        if 3 == self.dim:
            nbinsx = len(xval_ordered[0]) - 1
            binsx = array.array('d', xval_ordered[0])
            nbinsy = len(xval_ordered[1]) - 1
            binsy = array.array('d', xval_ordered[1])
            nbinsz = len(xval_ordered[2]) - 1
            binsz = array.array('d', xval_ordered[2])
            hist = self.get_hist_classes()[self.dim - 1](self.sanitize_name(name), '', nbinsx, binsx, nbinsy, binsy, nbinsz, binsz)

        for i in xrange(self.dim):
            name = self.independent_variables[i]['header']['name']
            if 'units' in self.independent_variables[i]['header']:
                name += ' [%s]' % self.independent_variables[i]['header']['units']
            name = unicode(name).encode('ascii', 'replace')
            getattr(hist, self._hist_axes_getters[i])().SetTitle(name)
            if 'labels' in self.independent_variables[i]: # set alphanumeric bin labels
                for ibin, label in enumerate(self.independent_variables[i]['labels']):
                    getattr(hist, self._hist_axes_getters[i])().SetBinLabel(ibin + 1, label)

        if self.dim < len(self.get_hist_classes()):
            name = self.dependent_variable['header']['name']
            if 'units' in self.dependent_variable['header']:
                name += ' [%s]' % self.dependent_variable['header']['units']
            name = unicode(name).encode('ascii', 'replace')
            getattr(hist, self._hist_axes_getters[self.dim])().SetTitle(name)

        for i in xrange(len(self.xval[0])):
            hist.Fill(*([self.xval[dim_i][i] for dim_i in xrange(self.dim)] + [self.yval[i]]))
        return hist

    def create_objects(self):
        self.calculate_total_errors()

        error_hists = []
        error_labels = {}
        error_indices = {}
        index = 0

        is_number_list = self.is_number_var(self.dependent_variable)

        for i, value in enumerate(self.dependent_variable.get('values', [])):

            if not is_number_list[i]: continue # skip non-numeric y values

            # process the labels to ensure uniqueness
            observed_error_labels = {}
            for error in value.get('errors', []):
                label = error.get('label', '')

                if label not in observed_error_labels:
                    observed_error_labels[label] = 0
                observed_error_labels[label] += 1

                if observed_error_labels[label] > 1:
                    error['label'] = label + '_' + str(observed_error_labels[label])

                # append "_1" to first error label that has a duplicate
                if observed_error_labels[label] == 2:
                    for error1 in value.get('errors', []):
                        error1_label = error1.get('label', 'error')
                        if error1_label == label:
                            error1['label'] = label + "_1"
                            break

            for error in value.get('errors', []):
                if 'label' not in error:
                    error['label'] = 'error'
                label = error['label']
                if label not in error_labels:
                    index += 1
                    error_indices[index] = label
                if 'symerror' in error and label not in error_labels:
                    error_labels[label] = 'symerror'
                elif 'asymerror' in error and error_labels.get(label, 'symerror') == 'symerror':
                    error_labels[label] = 'asymerror'

        yvals = []
        for index in xrange(1, len(error_labels) + 1):
            error_label = error_indices[index]
            if error_labels[error_label] == 'asymerror':
                yval_plus_label = error_label + '_plus'
                yval_plus = []
                yval_minus_label = error_label + '_minus'
                yval_minus = []

                for i, value in enumerate(self.dependent_variable.get('values', [])):
                    if not is_number_list[i]: continue # skip non-numeric y values
                    error = filter(lambda x: x.get('label') == error_label, value.get('errors', []))
                    if len(error) == 0:
                        yval_plus.append(0.0)
                        yval_minus.append(0.0)
                    elif 'symerror' in error[0]:
                        err_val = error_value_processor(value['value'], error[0]['symerror'])
                        yval_plus.append(err_val)
                        yval_minus.append(-err_val)
                    elif 'asymerror' in error[0]:
                        err_plus = error_value_processor(value['value'], error[0]['asymerror']['plus'])
                        err_min = error_value_processor(value['value'], error[0]['asymerror']['minus'])
                        yval_plus.append(err_plus)
                        yval_minus.append(err_min)
                    else:
                        yval_plus.append(0.0)
                        yval_minus.append(0.0)

                yvals += [(yval_plus_label, yval_plus, '%s%s' % (index, 'plus')),
                          (yval_minus_label, yval_minus, '%s%s' % (index, 'minus'))]
            else:
                yval = []

                for i, value in enumerate(self.dependent_variable.get('values', [])):
                    if not is_number_list[i]: continue # skip non-numeric y values
                    error = filter(lambda x: x.get('label') == error_label, value.get('errors', []))
                    if len(error) == 0:
                        yval.append(0.0)
                    elif 'symerror' in error[0]:
                        err_val = error_value_processor(value['value'], error[0]['symerror'])
                        yval.append(err_val)
                    else:
                        yval.append(0.0)

                yvals += [(error_label, yval, index)]

        for name, vals, index in yvals:
            try:
                error_hists.append(self._create_empty_hist(name, index, vals))
            except:
                log.error("Failed to create empty histogram")

        xval = []
        for i in xrange(self.dim):
            xval.append([])
            i_var = self.independent_variables[i]['values']
            for x in i_var:
                xval[i].append(x['low'])
            xval[i].append(i_var[-1]['high'])

        try:
            hist = self._create_hist(xval)
        except:
            log.error("Failed to create histogram")
            return [] + error_hists

        return [hist] + error_hists


class TH3FRootClass(THFRootClass):
    dim = 3


class TH2FRootClass(THFRootClass):
    dim = 2


class TH1FRootClass(THFRootClass):
    dim = 1


class TGraph2DErrorsClass(ObjectWrapper):
    @classmethod
    def match(cls, independent_variables_map, dependent_variable):
        if not super(TGraph2DErrorsClass, cls).match(independent_variables_map, dependent_variable):
            return False
        if len(independent_variables_map) == 2:
            for independent_variable in independent_variables_map:
                if not independent_variable['values']:
                    return False
            return True
        return False

    def create_objects(self):
        self.calculate_total_errors()

        # check that errors are symmetric (within a tolerance to allow for numerical rounding)
        tol = 1e-15
        if any([a - b > tol for a, b in zip(self.xerr_plus[0], self.xerr_minus[0])]) or \
           any([a - b > tol for a, b in zip(self.xerr_plus[1], self.xerr_minus[1])]) or \
           any([a - b > tol for a, b in zip(self.yerr_plus, self.yerr_minus)]):
            return []

        if len(self.xval[0]):
            graph = ROOTModule.TGraph2DErrors(len(self.xval[0]),
                                           array.array('d', self.xval[0]),
                                           array.array('d', self.xval[1]),
                                           array.array('d', self.yval),
                                           array.array('d', self.xerr_plus[0]),
                                           array.array('d', self.xerr_plus[1]),
                                           array.array('d', self.yerr_plus))
        else:
            return []

        graph.SetName("Graph2D_y%s" % (self.dependent_variable_index + 1))

        xname = self.independent_variables[0]['header']['name']
        if 'units' in self.independent_variables[0]['header']:
            xname += ' [%s]' % self.independent_variables[0]['header']['units']
        xname = unicode(xname).encode('ascii', 'replace')
        yname = self.independent_variables[1]['header']['name']
        if 'units' in self.independent_variables[1]['header']:
            yname += ' [%s]' % self.independent_variables[1]['header']['units']
        yname = unicode(yname).encode('ascii', 'replace')
        zname = self.dependent_variable['header']['name']
        if 'units' in self.dependent_variable['header']:
            zname += ' [%s]' % self.dependent_variable['header']['units']
        zname = unicode(zname).encode('ascii', 'replace')

        graph.GetXaxis().SetTitle(xname)
        graph.GetYaxis().SetTitle(yname)
        graph.GetZaxis().SetTitle(zname)

        return [graph]


class TGraphAsymmErrorsRootClass(ObjectWrapper):
    @classmethod
    def match(cls, independent_variables_map, dependent_variable):
        if not super(TGraphAsymmErrorsRootClass, cls).match(independent_variables_map, dependent_variable):
            return False
        if len(independent_variables_map) == 1:
            for independent_variable in independent_variables_map:
                if not independent_variable['values']:
                    return False
            return True
        return False

    def create_objects(self):
        self.calculate_total_errors()

        if len(self.xval[0]):
            graph = ROOTModule.TGraphAsymmErrors(len(self.xval[0]),
                                              array.array('d', self.xval[0]),
                                              array.array('d', self.yval),
                                              array.array('d', self.xerr_minus[0]),
                                              array.array('d', self.xerr_plus[0]),
                                              array.array('d', self.yerr_minus),
                                              array.array('d', self.yerr_plus))
        else:
            return []

        graph.SetName("Graph1D_y%s" % (self.dependent_variable_index + 1))

        xname = self.independent_variables[0]['header']['name']
        if 'units' in self.independent_variables[0]['header']:
            xname += ' [%s]' % self.independent_variables[0]['header']['units']
        xname = unicode(xname).encode('ascii', 'replace')
        yname = self.dependent_variable['header']['name']
        if 'units' in self.dependent_variable['header']:
            yname += ' [%s]' % self.dependent_variable['header']['units']
        yname = unicode(yname).encode('ascii', 'replace')
        graph.GetXaxis().SetTitle(xname)
        graph.GetYaxis().SetTitle(yname)

        return [graph]


class ROOT(ArrayWriter):
    help = 'Writes to ROOT format (binary) converts tables into files containing TH1 objects'
    class_list = [TH3FRootClass, TH2FRootClass, TH1FRootClass, TGraph2DErrorsClass, TGraphAsymmErrorsRootClass]

    def __init__(self, *args, **kwargs):
        super(ROOT, self).__init__(*args, **kwargs)
        self.extension = 'root'

    def _write_table(self, data_out, table):
        data_out.mkdir(table.name.replace('/','-'))
        data_out.cd(table.name.replace('/','-'))

        # if no independent variables, use bins of unit width and centred on integers (1, 2, 3, etc.)
        if not table.independent_variables and table.dependent_variables:
            if table.dependent_variables[0]['values']:
                table.independent_variables.append({'header': {'name': 'Bin number'}, 'values': []})
                for i, value in enumerate(table.dependent_variables[0]['values']):
                    table.independent_variables[0]['values'].append({'low': i+0.5, 'high': i+1.5})

        # if any non-numeric independent variable values, use bins of unit width and centred on integers (1, 2, 3, etc.)
        # store original variables as alphanumeric labels to be passed to ROOT histograms
        for ii, independent_variable in enumerate(table.independent_variables):
            if False in ObjectWrapper.is_number_var(independent_variable):
                independent_variable_bins = \
                    {'header': {'name': independent_variable['header']['name'] + ' bin'},
                     'values': [], 'labels': []}
                for i, value in enumerate(independent_variable['values']):
                    independent_variable_bins['values'].append({'low': i + 0.5, 'high': i + 1.5})
                    if 'value' in value:
                        independent_variable_bins['labels'].append(str(value['value']))
                    else:
                        independent_variable_bins['labels'].append(str(value['low']) + '-' + str(value['high']))
                table.independent_variables[ii] = independent_variable_bins

        if self.hepdata_doi:
            table_doi = 'doi:' + self.hepdata_doi + '/t' + str(table.index)
        else:
            table_doi = table.name
        f = ObjectFactory(self.class_list, table.independent_variables, table.dependent_variables)
        for graph in f.get_next_object():
            graph.SetTitle(table_doi)
            graph.Write()

    def _prepare_outputs(self, data_out, outputs):
        if isinstance(data_out, (str, unicode)):
            self.file_emulation = True
            outputs.append(ROOTModule.TFile.Open(data_out, 'RECREATE'))
        # multiple tables - require directory
        elif isinstance(data_out, ROOTModule.TFile):
            outputs.append(data_out)
        else:  # assume it's a file like object
            self.file_emulation = True
            filename = os.path.join(tempfile.mkdtemp(),'tmp.root')
            outputs.append(ROOTModule.TFile.Open(filename,'RECREATE'))

    def write(self, data_in, data_out, *args, **kwargs):
        """
        :param data_in:
        :type data_in: hepconverter.parsers.ParsedData
        :param data_out: filelike object
        :type data_out: file
        :param args:
        :param kwargs:
        """
        self._get_tables(data_in)

        self.file_emulation = False
        outputs = []
        self._prepare_outputs(data_out, outputs)
        output = outputs[0]
        for i in xrange(len(self.tables)):
            table = self.tables[i]

            self._write_table(output, table)

        if data_out != output and hasattr(data_out, 'write'):
            output.Flush()
            output.ReOpen('read')
            file_size = output.GetSize()
            buff = bytearray(file_size)
            output.ReadBuffer(buff, file_size)
            data_out.write(buff)

        if self.file_emulation:
            filename = output.GetName()
            output.Close()



import unittest

import datetime

from hepdata_converter.parsers import Parser
from hepdata_converter.parsers.oldhepdata_parser import OldHEPData
from hepdata_converter.parsers.yaml_parser import YAML
from hepdata_converter.testsuite import insert_paths


class ParserTestSuite(unittest.TestCase):
    """Test suite for Parser factory class
    """

    def test_get_specific_parser_oldhepdata(self):
        self.assertEqual(Parser.get_concrete_class('oldhepdata').__class__, OldHEPData.__class__)

    def test_get_specific_parser_nonexist(self):
        self.assertRaises(ValueError, Parser.get_concrete_class, 'nonexisting_parser')

    @insert_paths('yaml/ins1283183', 'yaml/ins1397637', 'yaml/ins699647', 'yaml/ins1413748')
    def test_parse_speed(self, test_submissions):
        _yaml_parser = YAML()
        for idx, test_submission in enumerate(test_submissions):
            start_time = datetime.datetime.now()
            _yaml_parser.parse(test_submission)
            end_time = datetime.datetime.now()
            print("Took {0}ms to parse {1}".format(end_time - start_time, test_submission))



import os
from random import randint
import tempfile
import unittest
import shutil
import time
import yaml


def _parse_path_arguments(sample_file_name):
    _sample_file_name = list(sample_file_name)
    sample_file_name = []
    for path_element in _sample_file_name:
        sample_file_name += path_element.split('/')

    return sample_file_name


def construct_testdata_path(path_elements):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'testdata', *path_elements)


class insert_data_as_file(object):
    def __init__(self, *sample_file_name):
        self.sample_file_name = _parse_path_arguments(sample_file_name)

    def __call__(self, function):
        def _inner(*args, **kwargs):
            args = list(args)
            with open(construct_testdata_path(self.sample_file_name), 'r') as f:
                args.append(f)
                function(*args, **kwargs)

        return _inner


class insert_data_as_str(object):
    def __init__(self, *sample_file_name):
        self.sample_file_name = _parse_path_arguments(sample_file_name)

    def __call__(self, function):
        def _inner(*args, **kwargs):
            with open(construct_testdata_path(self.sample_file_name), 'r') as f:
                data = f.read()
            args = list(args)
            args.append(data)
            function(*args, **kwargs)

        return _inner


class insert_path(object):
    def __init__(self, *sample_file_name):
        self.sample_file_name = _parse_path_arguments(sample_file_name)

    def __call__(self, function):
        def _inner(*args, **kwargs):
            args = list(args)
            args.append(construct_testdata_path(self.sample_file_name))
            function(*args, **kwargs)

        return _inner

class insert_paths(object):
    def __init__(self, *sample_file_name):
        self.sample_file_name = sample_file_name

    def __call__(self, function):
        def _inner(*args, **kwargs):
            args = list(args)
            directories = []
            for path in self.sample_file_name:
                directories.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'testdata', path))
            args.append(directories)
            function(*args, **kwargs)

        return _inner


class TMPDirMixin(object):
    def tearDown(self):
        shutil.rmtree(self.current_tmp)

    def setUp(self):
        self.current_tmp = os.path.join(tempfile.gettempdir(), str(int(time.time()))) + str(randint(0, 10000))
        try:
            os.mkdir(self.current_tmp)
        finally:
            pass


class ExtendedTestCase(unittest.TestCase):
    def assertMultiLineAlmostEqual(self, first, second, msg=None):
        if hasattr(first, 'readlines'):
            lines = first.readlines()
        elif isinstance(first, (str, unicode)):
            lines = first.split('\n')

        if hasattr(second, 'readlines'):
            orig_lines = second.readlines()
        elif isinstance(second, (str, unicode)):
            orig_lines = second.split('\n')

        self.assertEqual(len(lines), len(orig_lines))
        for i in xrange(len(lines)):
            self.assertEqual(lines[i].strip(), orig_lines[i].strip())

    def assertDirsEqual(self, first_dir, second_dir, file_content_parser=lambda x: list(yaml.load_all(x)), exclude=[], msg=None):
        self.assertEqual(list(os.walk(first_dir))[1:], list(os.walk(second_dir))[1:], msg)
        dirs = list(os.walk(first_dir))
        for file in dirs[0][2]:
            if file not in exclude:
                with open(os.path.join(first_dir, file)) as f1, open(os.path.join(second_dir, file)) as f2:
                    # separated into 2 variables to ease debugging if the need arises
                    d1 = file_content_parser(f1.read())
                    d2 = file_content_parser(f2.read())
                    self.assertEqual(d1, d2)



import abc
import unittest
from hepdata_converter.common import GetConcreteSubclassMixin


class ParserTestSuite(unittest.TestCase):
    """Test suite for Parser factory class
    """

    def test_get_all_subclasses(self):
        class A(GetConcreteSubclassMixin):
            pass

        class AB(A):
            pass

        class AC(A):
            pass

        class AAB(AB):
            __metaclass__ = abc.ABCMeta

            @abc.abstractmethod
            def abstract(self):
                pass

        self.assertEqual(set([AB, AC]), set(A.get_all_subclasses()))

        self.assertEqual(set([AB, AC, AAB]), set(A.get_all_subclasses(include_abstract=True)))


# -*- encoding: utf-8 -*-
import os
import hepdata_converter
from hepdata_converter.testsuite import insert_data_as_file, insert_path, insert_paths
from hepdata_converter.testsuite.test_writer import WriterTestSuite

__author__ = 'Michał Szostak'


class YODAWriterTestSuite(WriterTestSuite):
    @insert_path('yaml_full')
    @insert_data_as_file('yoda/full.yoda')
    def test_simple_parse(self, yaml_simple_path, yoda_template):
        output_file_path = os.path.join(self.current_tmp, 'datafile.yoda')
        hepdata_converter.convert(yaml_simple_path, output_file_path,
                                  options={'output_format': 'yoda'})

        self.assertNotEqual(os.stat(output_file_path).st_size, 0, 'output yoda file is empty')
        with open(output_file_path, 'r') as f:
            self.assertMultiLineAlmostEqual(f, yoda_template)

    @insert_paths('yaml/ins1283183', 'yaml/ins1397637', 'yaml/ins699647', 'yaml/ins1413748')
    def test_parse_all(self, test_submissions):

        for idx, test_submission in enumerate(test_submissions):
            output_file_path = os.path.join(self.current_tmp, 'data-{}.yoda'.format(idx))

            hepdata_converter.convert(test_submission, output_file_path,
                                      options={'output_format': 'yoda'})

            self.assertNotEqual(os.stat(output_file_path).st_size, 0, 'output yoda file is empty')



import os

import hepdata_converter
from hepdata_converter.parsers import yaml_parser
from hepdata_converter.parsers.oldhepdata_parser import OldHEPData
from hepdata_converter.testsuite import insert_data_as_file, insert_path
from hepdata_converter.testsuite.test_writer import WriterTestSuite
from hepdata_converter.writers.yaml_writer import YAML


class OldHEPDataTestSuite(WriterTestSuite):
    """Test suite for OldHEPData parser (and YAML writer).
    These are mostly class tests, and unit tests
    """

    @insert_data_as_file('oldhepdata/sample.input')
    @insert_path('oldhepdata/yaml')
    def test_parse_submission(self, oldhepdata_file, yaml_path):
        oldhepdata_p = OldHEPData()
        oldhepdata_parsed_data = oldhepdata_p.parse(oldhepdata_file)
        yaml_p = yaml_parser.YAML()
        yaml_parsed_data = yaml_p.parse(yaml_path)

        self.assertEqual(yaml_parsed_data, oldhepdata_parsed_data)

    @insert_path('oldhepdata/sample.input')
    @insert_path('oldhepdata/yaml')
    def test_cli(self, oldhepdata_path, oldhepdata_yaml_path):
        hepdata_converter._main(
                ['--input-format', 'oldhepdata', '--output-format', 'yaml',
                 oldhepdata_path, self.current_tmp])

        self.assertDirsEqual(oldhepdata_yaml_path, self.current_tmp)

    @insert_data_as_file('oldhepdata/1396331.oldhepdata')
    @insert_path('oldhepdata/1396331-yaml')
    def test_parse_large_old_submission(self, oldhepdata_file, yaml_path):
        print('Testing on oldhepdata/1396331.oldhepdata')
        oldhepdata_p = OldHEPData()
        oldhepdata_parsed_data = oldhepdata_p.parse(oldhepdata_file)

        yaml_p = yaml_parser.YAML()
        yaml_parsed_data = yaml_p.parse(yaml_path)

        assert(len(yaml_parsed_data.tables) is len(oldhepdata_parsed_data.tables))



import os
import hepdata_converter
from hepdata_converter.testsuite import insert_data_as_str, insert_path
from hepdata_converter.testsuite.test_writer import WriterTestSuite


class CLIToolsTestSuite(WriterTestSuite):
    def test_wrong_call(self):
        self.assertRaises(SystemExit, hepdata_converter.main, [])

    @insert_path('yaml_full')
    @insert_data_as_str('csv/table_1.csv')
    def test_convert_yaml2csv(self, submission_file, table_csv):
        output_path = os.path.join(self.current_tmp, 'output.csv')
        code, message = hepdata_converter._main(['--input-format', 'yaml', '--output-format', 'csv',
                                                 '--table', 'Table 1', '--pack', submission_file, output_path])

        self.assertEqual(code, 0, message)
        self.assertTrue(os.path.exists(output_path))

        with open(output_path) as f:
            self.assertMultiLineAlmostEqual(table_csv, f.read())

    def test_convert_yaml2yoda(self):
        hepdata_converter._main(['--input-format', 'yaml', '--output-format', 'csv',
                                '--table', 'Table 1',
                                self.current_tmp, os.path.join(self.current_tmp, 'output.csv')])

    def test_help(self):
        self.assertRaises(SystemExit, hepdata_converter.main, ['--help'])

    def test_version(self):
        r, message = hepdata_converter._main(['--version'])
        self.assertEqual(r, 0)
        self.assertTrue(message.endswith(hepdata_converter.version.__version__))



import StringIO
import os
import yaml
import hepdata_converter
from hepdata_converter.testsuite import insert_path
from hepdata_converter.testsuite.test_writer import WriterTestSuite


class ConvertTestSuite(WriterTestSuite):
    """Test suite for Parser factory class
    """

    def setUp(self):
        super(ConvertTestSuite, self).setUp()
        self.simple_submission = (
            "*reference: ARXIV:1211.6096 : 2012 # arXiv number\n"
            "*reference: CERN-PH-EP-2012-318 : 2012 # preprint number\n"
            "*reference: JHEP 1303,128 (2013) : 2013 # journal reference\n"
            "*comment: CERN-LHC.  Measurements of the cross section for ZZ ...\n"
            "*dataset:\n"
            "*location: Page 17 of preprint\n"
            "*dscomment: The measured fiducial cross sections.  The first systematic uncertainty is the combined systematic uncertainty excluding luminosity, the second is the luminosity\n"
            "*reackey: P P --> Z0 Z0 X\n"
            "*obskey: SIG\n"
            "*qual: RE : P P --> Z0 < LEPTON+ LEPTON- > Z0 < LEPTON+ LEPTON- > X : P P --> Z0 < LEPTON+ LEPTON- > Z0* < LEPTON+ LEPTON- > X : P P --> Z0 < LEPTON+ LEPTON- > Z0 < NU NUBAR > X\n"
            "*xheader: SQRT(S) IN GEV\n"
            "*yheader: SIG(fiducial) IN FB\n"
            "*data: x : y : y : y\n"
            "7000; 25.4 +3.3,-3.0 (DSYS=+1.2,-1.0,DSYS=1.0:lumi); 29.8 +3.8,-3.5 (DSYS=+1.7,-1.5,DSYS=1.2:lumi); 12.7 +3.1,-2.9 (DSYS=1.7,DSYS=0.5:lumi);\n"
            "*dataend:\n"
        )

        self.correct_submit_output = (
            "additional_resources: []\n"
            "comment: 'CERN-LHC.  Measurements of the cross section for ZZ ...'\n"
            "---\n"
            "additional_resources: []\n"
            "data_file: data1.yaml\n"
            "data_license: {description: null, name: null, url: null}\n"
            "description: 'The measured fiducial cross sections.  The first systematic uncertainty\n"
            "  is the combined systematic uncertainty excluding luminosity, the second is the luminosity.'\n"
            "keywords:\n"
            "- name: reactions\n"
            "  values: ['P P --> Z0 Z0 X']\n"
            "- name: observables\n"
            "  values: ['SIG']\n"
            "- name: phrases\n"
            "  values: []\n"
            "- name: cmenergies\n"
            "  values: [7000.0]\n"
            "location: 'Data from Page 17 of preprint'\n"
            "name: Table 1\n"
        )

        self.correct_table_output = (
            "dependent_variables:\n"
            "- header: {name: SIG(fiducial), units: FB}\n"
            "  qualifiers:\n"
            "  - {name: RE, value: P P --> Z0 < LEPTON+ LEPTON- > Z0 < LEPTON+ LEPTON- > X}\n"
            "  values:\n"
            "  - errors:\n"
            "    - asymerror: {minus: '-3.0', plus: '3.3'}\n"
            "      label: stat\n"
            "    - asymerror: {minus: '-1.0', plus: '1.2'}\n"
            "      label: sys\n"
            "    - {label: 'sys,lumi', symerror: '1.0'}\n"
            "    value: '25.4'\n"
            "- header: {name: SIG(fiducial), units: FB}\n"
            "  qualifiers:\n"
            "  - {name: RE, value: P P --> Z0 < LEPTON+ LEPTON- > Z0* < LEPTON+ LEPTON- > X}\n"
            "  values:\n"
            "  - errors:\n"
            "    - asymerror: {minus: '-3.5', plus: '3.8'}\n"
            "      label: stat\n"
            "    - asymerror: {minus: '-1.5', plus: '1.7'}\n"
            "      label: sys\n"
            "    - {label: 'sys,lumi', symerror: '1.2'}\n"
            "    value: '29.8'\n"
            "- header: {name: SIG(fiducial), units: FB}\n"
            "  qualifiers:\n"
            "  - {name: RE, value: P P --> Z0 < LEPTON+ LEPTON- > Z0 < NU NUBAR > X}\n"
            "  values:\n"
            "  - errors:\n"
            "    - asymerror: {minus: '-2.9', plus: '3.1'}\n"
            "      label: stat\n"
            "    - {label: sys, symerror: '1.7'}\n"
            "    - {label: 'sys,lumi', symerror: '0.5'}\n"
            "    value: '12.7'\n"
            "independent_variables:\n"
            "- header: {name: SQRT(S), units: GEV}\n"
            "  values: ['7000']\n"
        )

    def test_bad_argumets(self):
        self.assertRaises(ValueError, hepdata_converter.convert, None, None)

    def test_submission(self):
        hepdata_converter.convert(StringIO.StringIO(self.simple_submission), self.current_tmp, options={'input_format': 'oldhepdata'})

        with open(os.path.join(self.current_tmp, 'submission.yaml')) as submission_file:
            self.assertEqual(list(yaml.load_all(submission_file)), list(yaml.load_all(self.correct_submit_output)))

    def test_not_implemented_writer(self):
        """This feature is not implemented yet, but to get test coverage it is tested,
        when this functionality is implemented this test case should either be modified or deleted
        """
        self.assertRaises(ValueError, hepdata_converter.convert, None, None, options={'input_format': 'yaml', 'output_format': 'not_implemented'})

    @insert_path('yaml_full')
    def test_same_type_conversion(self, yaml_path):
        hepdata_converter.convert(yaml_path, self.current_tmp, options={'input_format': 'yaml', 'output_format': 'yaml'})
        # exclude data6.yaml and data7.yaml because they are not listed in submission.yaml
        self.assertDirsEqual(yaml_path, self.current_tmp, exclude=['data6.yaml', 'data7.yaml'])


import abc
import copy
import os
from hepdata_converter.common import GetConcreteSubclassMixin, OptionInitMixin

__all__ = []

import pkgutil
import inspect


class BadFormat(Exception):
    """Class for exceptions raised if bad formatting of parser's input file prohibits from
    parsing the file correctly

    """
    pass


class Table(object):
    """Simple table storage class used for handling tables without extremely deep dictionaries
    """
    def __eq__(self, other):
        if self.metadata != other.metadata:
            return False
        if self.data != other.data:
            return False
        if self.all_data != other.all_data:
            return False
        if self.index != other.index:
            return False
        return True

    def __init__(self, index=None, data_file=None, table_name=None, metadata=None, data=None):
        self.data = data or []
        self.index = index
        self.qualifiers = []
        self.dserrors = []

        self.metadata = metadata or {
            'name': table_name or 'Table %s' % self.index,
            'location': '',
            'description': '',
            'keywords': [
                {'name': 'reactions', 'values': []},
                {'name': 'observables', 'values': []},
                {'name': 'phrases', 'values': []},
                {'name': 'cmenergies', 'values': []},
            ],
            'data_file': data_file or 'data%s.yaml' % index,
            # it seems it's required
            # TODO - is it really required? should sensible defaults be provided?
            'data_license': {
                'name': None,
                'url': None,
                'description': None # (optional)
            },
            'additional_resources': [
                # Below in the comments are listed allowed keys / values for additional_resources
                #
                # {
                # 'location': None
                # 'description': "Full source code for creating this data"
                # },
                # {'location': "root.root",
                #  'description': "Some file",
                #  'license': {
                #     'name': 'GPL 2',
                #     'url': 'url for license',
                #     'description': 'Tell me about it. This can appear in the main record display' # (optional)
                #     }
                # }
            ]
        }
        for keyword in self.metadata['keywords']:
            if keyword['name'] == 'reactions':
                self.reactions = keyword['values']
            elif keyword['name'] == 'observables':
                self.observables = keyword['values']
            elif keyword['name'] == 'phrases':
                self.phrases = keyword['values']
            elif keyword['name'] == 'cmenergies':
                self.energies = keyword['values']

        self.data_file = self.metadata['data_file']

        if self.data:
            for dependent_variable in self.data.get('dependent_variables', []):
                self.qualifiers.append(dependent_variable.get('qualifiers', []))

    @property
    def name(self):
        return self.metadata['name']

    @property
    def additional_resources(self):
        return self.metadata['additional_resources']

    @property
    def data_license(self):
        return self.metadata['data_license']

    @property
    def keywords(self):
        return self.metadata['keywords']

    @property
    def location(self):
        return self.metadata['location']

    @property
    def description(self):
        return self.metadata['description']

    @property
    def independent_variables(self):
        return self.data['independent_variables']

    @property
    def dependent_variables(self):
        return self.data['dependent_variables']

    @property
    def all_data(self):
        _all = copy.copy(self.metadata)
        if 'data_file' in self.metadata:
            del _all['data_file']

        _all['independent_variables'] = self.independent_variables
        _all['dependent_variables'] = self.dependent_variables

        return _all

class ParsedData(object):
    """Simple data storage class which should be returned by Parser.parse method

    """
    def __init__(self, data, tables):
        self.data = data
        self.tables = tables

    def __eq__(self, other):
        if self.data != other.data:
            return False
        if self.tables != other.tables:
            return False
        return True

    def get_table(self, **kwargs):
        assert len(kwargs) == 1
        key, search_val = kwargs.items()[0]
        assert key in ('id', 'name', 'file')

        if key == 'id':
            return self.tables[search_val]
        elif key == 'file':
            for table in self.tables:
                if table.data_file == os.path.basename(search_val):
                    return table
            raise IndexError("No table with filename = %s" % search_val)
        elif key == 'name':
            for table in self.tables:
                if table.name == search_val:
                    return table
            raise IndexError("No table with name = %s" % search_val)


class Parser(GetConcreteSubclassMixin, OptionInitMixin):
    __metaclass__  = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        OptionInitMixin.__init__(self, options=kwargs)

    @abc.abstractmethod
    def parse(self, data_in, *args, **kwargs):
        """Parses data_in and returns parsed data in the form of ParsedData object

        :param data_in: data to process, depending on the concrete implementation object type may vary
        :param options: dictionary containing additional options for concrete Parser

        :return: Parsed data, in the YAML writer friendly format (list, dicts, and simple objects)
        :rtype: ParsedData
        """

# import all packages in the parsers package, so that Parser.get_specific_parser will recognise them
for loader, name, is_pkg in pkgutil.walk_packages(__path__):
    module = loader.find_module(name).load_module(name)

    for name, value in inspect.getmembers(module):
        if name.startswith('__'):
            continue

        globals()[name] = value
        __all__.append(name)



import yaml
# We try to dump using the CSafeDumper for speed improvements.
try:
    from yaml import CSafeDumper as Dumper
except ImportError: #pragma: no cover
    from yaml import SafeDumper as Dumper #pragma: no cover
from hepdata_converter.common import Option, OptionInitMixin
from hepdata_converter.writers import Writer
import os

def str_presenter(dumper, data):
    if '\n' in data:
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)

class YAML(Writer):
    help = 'Writes YAML output. Output should be defined as filepath to the directory where submission.yaml and associated ' \
           'table files will be written'

    @classmethod
    def options(cls):
        options = Writer.options()
        options['single_file'] = Option('single-file', type=bool, default=False, variable_mapping='single_file',
                                        required=False, help="If set output will be written to single yaml file, instead "
                                                             "of multiple files (separating data and metadata of the tables)")
        return options

    def __init__(self, *args, **kwargs):
        super(YAML, self).__init__(single_file_output=True, *args, **kwargs)
        Dumper.add_representer(str, str_presenter)

    def write(self, data_in, data_out, *args, **kwargs):
        """

        :param data_in:
        :type data_in: hepconverter.parsers.ParsedData
        :param data_out: path of the directory to which yaml files will be written
        :type data_out: str
        :param args:
        :param kwargs:
        """

        tables = data_in.tables
        data = data_in.data

        if self.hepdata_doi:
            data['hepdata_doi'] = self.hepdata_doi
            for table in tables:
                table.metadata['table_doi'] = self.hepdata_doi + '/t' + str(table.index)

        if not isinstance(data_out, (str, unicode)) and not self.single_file:
            raise ValueError("output is not string, and single_file flag is not specified")

        if not self.single_file:
            self.create_dir(data_out)
            with open(os.path.join(data_out, 'submission.yaml'), 'w') as submission_file:
                yaml.dump_all([data] + [table.metadata for table in tables], submission_file, Dumper=Dumper)

                for table in tables:
                    with open(os.path.join(data_out, table.data_file), 'w') as table_file:
                        yaml.dump(table.data, table_file, Dumper=Dumper)
        else:
            if isinstance(data_out, (str, unicode)):
                with open(data_out, 'w') as submission_file:
                    yaml.dump_all([data] + [table.all_data for table in tables], submission_file, Dumper=Dumper)
            else: # expect filelike object
                yaml.dump_all([data] + [table.all_data for table in tables], data_out, Dumper=Dumper)


from hepdata_converter.common import Option
from hepdata_converter.writers.array_writer import ArrayWriter, ObjectWrapper, ObjectFactory
import yoda


class ScatterYodaClass(ObjectWrapper):
    dim = -1
    _scatter_classes = [yoda.core.Scatter1D, yoda.core.Scatter2D, yoda.core.Scatter3D]
    _point_classes = [yoda.core.Point1D, yoda.core.Point2D, yoda.core.Point3D]

    @classmethod
    def get_scatter_cls(cls):
        return cls._scatter_classes[cls.dim]

    @classmethod
    def get_point_cls(cls):
        return cls._point_classes[cls.dim]

    @classmethod
    def match(cls, independent_variables_map, dependent_variable):
        if not ObjectWrapper.match(independent_variables_map, dependent_variable):
            return False
        elif len(independent_variables_map) == cls.dim:
            for independent_variable in independent_variables_map:
                if not independent_variable['values']:
                    return False
            return True
        return False

    def _create_scatter(self):
        graph = self.get_scatter_cls()()

        for i in xrange(len(self.yval)):
            args = []

            for dim_i in xrange(self.dim):
                args.append(self.xval[dim_i][i])
            args.append(self.yval[i])
            for dim_i in xrange(self.dim):
                args.append([self.xerr_minus[dim_i][i], self.xerr_plus[dim_i][i]])
            args.append([self.yerr_minus[i], self.yerr_plus[i]])

            graph.addPoint(self.get_point_cls()(*args))

        return graph

    def create_objects(self):
        self.calculate_total_errors()

        graph = self._create_scatter()

        return [graph]


class Scatter1DYodaClass(ScatterYodaClass):
    dim = 0


class Scatter2DYodaClass(ScatterYodaClass):
    dim = 1


class Scatter3DYodaClass(ScatterYodaClass):
    dim = 2


class YODA(ArrayWriter):
    help = 'Writes YODA output for table specified by --table parameter, the output should be defined as ' \
           'filepath to output yoda file'

    class_list = [Scatter3DYodaClass, Scatter2DYodaClass, Scatter1DYodaClass]

    @classmethod
    def options(cls):
        options = ArrayWriter.options()
        options['rivet_analysis_name'] = Option('rivet-analysis-name', 'r', type=str, default='RIVET_ANALYSIS_NAME',
                                                required=False, variable_mapping='rivet_analysis_name',
                                                help='Rivet analysis name, e.g. "ATLAS_2016_I1424838"')
        return options

    def __init__(self, *args, **kwargs):
        super(YODA, self).__init__(*args, **kwargs)
        self.extension = 'yoda'

    def _prepare_outputs(self, data_out, outputs):
        if isinstance(data_out, (str, unicode)):
            self.file_emulation = True
            outputs.append(open(data_out, 'w'))
        # multiple tables - require directory
        else:  # assume it's a file like object
            self.file_emulation = True
            outputs.append(data_out)

    def _write_table(self, data_out, table):
        # if any non-numeric independent variables, use bins of unit width and centred on integers (1, 2, 3, etc.)
        for ii, independent_variable in enumerate(table.independent_variables):
            if False in ObjectWrapper.is_number_var(independent_variable):
                for i, value in enumerate(independent_variable['values']):
                    table.independent_variables[ii]['values'][i] = {'low': i + 0.5, 'high': i + 1.5}
        table_num = str(table.index)
        if self.hepdata_doi:
            table_doi = 'doi:' + self.hepdata_doi + '/t' + table_num
        else:
            table_doi = table.name
        f = ObjectFactory(self.class_list, table.independent_variables, table.dependent_variables)
        for idep, graph in enumerate(f.get_next_object()):
            graph.title = table_doi
            graph.path = '/REF/' + self.rivet_analysis_name + '/' \
                         + 'd' + table_num.zfill(2) + '-x01-y' + str(idep + 1).zfill(2)
            graph.setAnnotation('IsRef', '1')
            yoda.core.writeYODA(graph, data_out)

    def write(self, data_in, data_out, *args, **kwargs):
        """

        :param data_in:
        :type data_in: hepconverter.parsers.ParsedData
        :param data_out: filelike object
        :type data_out: file
        :param args:
        :param kwargs:
        """
        self._get_tables(data_in)

        self.file_emulation = False
        outputs = []
        self._prepare_outputs(data_out, outputs)
        output = outputs[0]

        for i in xrange(len(self.tables)):
            table = self.tables[i]

            self._write_table(output, table)

        if self.file_emulation:
            output.close()



# -*- encoding: utf-8 -*-
import os
import hepdata_converter
from hepdata_converter.testsuite import insert_path, insert_paths
from hepdata_converter.testsuite.test_writer import WriterTestSuite
import ROOT

__author__ = 'Michał Szostak'

def walk(tdirectory,
         path=None,
         depth=0):
    '''Walk the directory structure recursively and return its content.
    '''
    dirnames, objectnames = [], []
    for key in list(tdirectory.GetListOfKeys()):
        name = key.GetName()
        classname = key.GetClassName()
        is_directory = classname.startswith('TDirectory')
        if is_directory:
            dirnames.append(name)
        else:
            objectnames.append(name)
    if path:
        dirpath = os.path.join(path, tdirectory.GetName())
    elif not isinstance(tdirectory, ROOT.TFile):
        dirpath = tdirectory.GetName()
    else:
        dirpath = ''
    yield dirpath, dirnames, objectnames
    for dirname in dirnames:
        rdir = tdirectory.GetDirectory(dirname)
        for x in walk(rdir,
                      depth=depth + 1,
                      path=dirpath):
            yield x

class ROOTWriterTestSuite(WriterTestSuite):


    @insert_path('yaml_full')
    @insert_path('root/full.root')
    def test_simple_parse(self, yaml_full_path, full_root_path):
        output_file_path = os.path.join(self.current_tmp, 'datafile.root')
        hepdata_converter.convert(yaml_full_path, output_file_path,
                                  options={'output_format': 'root'})

        self.assertNotEqual(os.stat(output_file_path).st_size, 0, 'output root file is empty')

        f = ROOT.TFile.Open(output_file_path)
        f_orig = ROOT.TFile.Open(full_root_path)
        self.assertEqual(list(walk(f)), list(walk(f_orig)))
        f.Close()
        f_orig.Close()

        with open(output_file_path, 'w') as output:
            hepdata_converter.convert(yaml_full_path, output,
                                      options={'output_format': 'root'})

        self.assertNotEqual(os.stat(output_file_path).st_size, 0, 'output root file is empty')

        f = ROOT.TFile.Open(output_file_path)
        f_orig = ROOT.TFile.Open(full_root_path)
        self.assertEqual(list(walk(f)), list(walk(f_orig)))
        for path, dirs, objects in list(walk(f))[1:]:
            for obj in objects:
                o = f.Get('%s/%s' % (path, obj))
                o_orig = f_orig.Get('%s/%s' % (path, obj))
                self.assertEqual(o.__class__, o_orig.__class__)
                if o.__class__.__name__.startswith('TGraph'):
                    self.assertEqual(o.GetN(), o_orig.GetN())
                    for i in xrange(o.GetN()):
                        self.assertEqual(o.GetX()[i],o_orig.GetX()[i])
                        self.assertEqual(o.GetY()[i],o_orig.GetY()[i])
        f.Close()
        f_orig.Close()

    @insert_path('yaml_full')
    @insert_path('root/full.root')
    def test_th1_parse(self, yaml_full_path, full_root_path):
        output_file_path = os.path.join(self.current_tmp, 'datafile.root')
        hepdata_converter.convert(yaml_full_path, output_file_path,
                                  options={'output_format': 'root', 'table': 'data2.yaml'})
        pass

    @insert_paths('yaml/ins1283183', 'yaml/ins1397637', 'yaml/ins699647', 'yaml/ins1413748')
    def test_parse_all(self, test_submissions):

        for idx, test_submission in enumerate(test_submissions):
            output_file_path = os.path.join(self.current_tmp, 'data-{}.root'.format(idx))
            hepdata_converter.convert(test_submission, output_file_path,
                                      options={'output_format': 'root'})

            self.assertNotEqual(os.stat(output_file_path).st_size, 0, 'output root file is empty')



import yaml
# We try to load using the CSafeLoader for speed improvements.
try:
    from yaml import CSafeLoader as Loader
except ImportError: #pragma: no cover
    from yaml import SafeLoader as Loader #pragma: no cover
from hepdata_validator.submission_file_validator import SubmissionFileValidator
from hepdata_validator.data_file_validator import DataFileValidator
from hepdata_converter.parsers import Parser, ParsedData, Table
from multiprocessing import Pool
import os, re

# Allow for a bug in PyYAML where numbers like 1e+04 are parsed as strings not as floats.
# See https://stackoverflow.com/a/30462009
# Try replacing PyYAML by ruamel.yaml in future?
Loader.add_implicit_resolver(
    u'tag:yaml.org,2002:float',
    re.compile(u'''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$''', re.X),
    list(u'-+0123456789.'))

class YAML(Parser):
    help = 'Parses New HEPData YAML format. Input parameter should be path to ' \
           'the directory where submission.yaml file ' \
           'is present (or direct filepath to the submission.yaml file)'

    pool = Pool()

    def __init__(self, *args, **kwargs):
        super(YAML, self).__init__(*args, **kwargs)

    def _pretty_print_errors(self, message_dict):
        return ' '.join(
                ['%s: %s' % (key, ' | '.join([e.message for e in val])) for
                 key, val in message_dict.items()])

    def parse(self, data_in, *args, **kwargs):
        """
        :param data_in: path to submission.yaml
        :param args:
        :param kwargs:
        :raise ValueError:
        """
        if not os.path.exists(data_in):
            raise ValueError("File / Directory does not exist: %s" % data_in)

        if os.path.isdir(data_in):
            submission_filepath = os.path.join(data_in, 'submission.yaml')
            if not os.path.exists(submission_filepath):
                submission_filepath = os.path.join(data_in, 'submission.yml')
                if not os.path.exists(submission_filepath):
                    raise ValueError("No submission file in %s" % data_in)
            data_in = submission_filepath

        # first validate submission file:
        with open(data_in, 'r') as submission_file:
            submission_data = list(yaml.load_all(submission_file, Loader=Loader))

            if len(submission_data) == 0:
                raise RuntimeError("Submission file (%s) is empty" % data_in)

            submission_file_validator = SubmissionFileValidator()
            if not submission_file_validator.validate(file_path=data_in,
                                                      data=submission_data):
                raise RuntimeError(
                    "Submission file (%s) did not pass validation: %s" %
                    (data_in, self._pretty_print_errors(
                        submission_file_validator.get_messages())))

        metadata = {}
        tables = []

        # validator for table data
        data_file_validator = DataFileValidator()

        index = 0
        for i in range(0, len(submission_data)):
            if not submission_data[i]: # empty YAML document
                continue
            if 'data_file' not in submission_data[i]:
                metadata = submission_data[i] # information about whole submission
                continue
            table_filepath = os.path.join(os.path.dirname(data_in),
                                          submission_data[i]['data_file'])
            with open(table_filepath, 'r') as table_file:
                if not os.path.exists(table_filepath):
                    raise ValueError(
                        "table file: %s does not exist" % table.data_file)

                table_data = yaml.load(table_file, Loader=Loader)

                if not data_file_validator.validate(data=table_data,
                                                    file_path=table_filepath):
                    raise RuntimeError(
                        "Data file (%s) did not pass validation: %s" %
                        (table_filepath, self._pretty_print_errors(
                            data_file_validator.get_messages())))

                index = index + 1
                table = Table(index=index, metadata=submission_data[i],
                              data=table_data)
                tables.append(table)

        return ParsedData(metadata, tables)
